from troposphere.dynamodb import Table as DynamoDBTable, PointInTimeRecoverySpecification, SSESpecification

from serverless.aws.iam.dynamodb import (
    DynamoDBFullAccess,
    DynamoDBReader,
    DynamoDBWriter,
    DynamoDBDelete,
    DynamoDBWriteOnly,
)
from . import Resource
from .kms import EncryptableResource
from ..features.encryption import Encryption
from ..iam import PolicyBuilder, IAMPreset
from ...service import Identifier


class Table(Resource):
    def __init__(self, TableName, with_full_access=False, with_read_access=False, **kwargs):
        if "${sls:stage}" not in TableName:
            TableName += "-${sls:stage}"

        kwargs.setdefault(
            "PointInTimeRecoverySpecification", PointInTimeRecoverySpecification(PointInTimeRecoveryEnabled=True)
        )

        kwargs.setdefault("DeletionPolicy", "Retain")

        super().__init__(
            DynamoDBTable(title=TableName.replace("${sls:stage}", "").strip("-"), TableName=TableName, **kwargs)
        )
        self.access = None

        if with_full_access:
            self.with_read_access()

        if with_read_access:
            self.with_read_access()

    def configure(self, service):
        if service.has(Encryption):
            self.resource.SSESpecification = SSESpecification(
                KMSMasterKeyId=EncryptableResource.encryption_key(), SSEEnabled=True, SSEType="KMS"
            )

        if service.service.pascal not in self.resource.TableName:
            self.resource.TableName = service.service.pascal + self.resource.TableName

    def _apply(self, preset: IAMPreset, builder: PolicyBuilder = None):
        if builder:
            builder.apply(preset)
        else:
            self.access = preset

        return self

    def with_full_access(self, builder: PolicyBuilder = None):
        return self._apply(DynamoDBFullAccess(self.resource), builder)

    def with_read_access(self, builder: PolicyBuilder = None):
        return self._apply(DynamoDBReader(self.resource), builder)

    def with_write_access(self, builder: PolicyBuilder = None):
        return self._apply(DynamoDBWriter(self.resource), builder)

    def enable_read(self, builder: PolicyBuilder):
        return self.with_read_access(builder)

    def enable_write(self, builder: PolicyBuilder):
        return self._apply(DynamoDBWriteOnly(self.resource), builder)

    def enable_delete(self, builder: PolicyBuilder):
        return self._apply(DynamoDBDelete(self.resource), builder)

    @property
    def table_arn(self):
        return self.resource.Ref().to_dict()

    def variables(self):
        return {
            "TABLE_"
            + Identifier(self.resource.TableName.replace("-${sls:stage}", "").replace("${sls:stage}", "")).snake.upper(): self.resource.TableName
        }

    def resources(self):
        return [self.resource]

    def permissions(self):
        if self.access:
            return [self.access]
        return []



from serverless.service.plugins.generic import Generic


class ComposedVars(Generic):
    yaml_tag = "!ComposedVarsPlugin"

    def __init__(self):
        super().__init__("serverless-plugin-composed-vars")



from serverless.aws.functions.generic import Function
from serverless.service import YamlOrderedDict


class S3Event(YamlOrderedDict):
    yaml_tag = "s3"

    def __init__(self, bucket, event, rules=None):
        super().__init__()
        self.bucket = bucket
        self.event = event

        if rules:
            self.rules = rules


class S3Function(Function):
    yaml_tag = "!S3Function"

    def __init__(
        self,
        service,
        name,
        description,
        bucket,
        event,
        rules=None,
        handler=None,
        timeout=None,
        layers=None,
        use_dlq=True,
        use_async_dlq=True,
        **kwargs,
    ):
        super().__init__(
            service, name, description, handler, timeout, layers, use_dlq=use_dlq, use_async_dlq=use_async_dlq, **kwargs
        )
        self.trigger(S3Event(bucket, event, rules))



from serverless.aws.functions.generic import Function
from serverless.service.types import YamlOrderedDict


class KinesisEvent(YamlOrderedDict):
    yaml_tag = "stream"

    def __init__(
        self,
        arn,
        batchSize=10,
        maximumRecordAgeInSeconds=120,
        maximumRetryAttempts=5,
        startingPosition="LATEST",
        functionResponseType="ReportBatchItemFailures",
        destinations=None,
    ):
        super().__init__()
        self.arn = arn
        self.batchSize = batchSize
        self.maximumRecordAgeInSeconds = maximumRecordAgeInSeconds
        self.maximumRetryAttempts = maximumRetryAttempts
        self.startingPosition = startingPosition
        self.functionResponseType = functionResponseType

        if destinations:
            self.destinations = destinations


class KinesisFunction(Function):
    yaml_tag = "!KinesisFunction"

    def __init__(
        self,
        stream,
        service,
        name,
        description,
        handler=None,
        timeout=None,
        layers=None,
        batch_size=10,
        maximum_record_age_in_seconds=120,
        maximum_retry_attempts=5,
        starting_position="LATEST",
        function_response_type="ReportBatchItemFailures",
        destinations=None,
        use_dlq=True,
        use_async_dlq=True,
        **kwargs,
    ):
        super().__init__(
            service, name, description, handler, timeout, layers, use_dlq=use_dlq, use_async_dlq=use_async_dlq, **kwargs
        )
        self.trigger(
            KinesisEvent(
                arn=stream,
                batchSize=batch_size,
                maximumRecordAgeInSeconds=maximum_record_age_in_seconds,
                maximumRetryAttempts=maximum_retry_attempts,
                startingPosition=starting_position,
                functionResponseType=function_response_type,
                destinations=destinations,
            )
        )



from serverless.service.plugins.generic import Generic


class PythonRequirements(Generic):
    yaml_tag = "!PythonRequirementsPlugin"

    def __init__(
        self,
        dockerizePip=True,
        usePoetry=True,
        useDownloadCache=True,
        useStaticCache=True,
        layer=True,
        dockerImage=None,
        pipCmdExtraArgs=None,
        dockerSsh=False,
    ):
        super().__init__("serverless-python-requirements")

        self.dockerImage = dockerImage
        self.dockerizePip = dockerizePip
        self.usePoetry = usePoetry
        self.useDownloadCache = useDownloadCache
        self.useStaticCache = useStaticCache
        self.layer = layer
        self.pipCmdExtraArgs = pipCmdExtraArgs or []
        self.dockerSsh = dockerSsh

    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        if not export.get("dockerImage"):
            export["dockerImage"] = f"lambci/lambda:build-{service.provider.runtime}"

        service.custom.pythonRequirements = export



from typing import Dict, Optional

from serverless.service.plugins.generic import Generic


class InvalidInputException(Exception):
    pass


class DeploymentBucket(Generic):
    """
    Plugin homepage: https://www.npmjs.com/package/serverless-deployment-bucket
    Example of use
    ```python
    service.plugins.add(DeploymentBucket())
    ```
    """

    yaml_tag = "!DeploymentBucketPlugin"

    def __init__(
        self,
        name: Optional[str] = None,
        serverSideEncryption: Optional[str] = None,
        kmsKeyID: Optional[str] = None,
        versioning: Optional[bool] = None,
        accelerate: Optional[bool] = None,
        blockPublicAccess: Optional[bool] = None,
        tags: Optional[Dict] = None,
    ):
        super().__init__("serverless-deployment-bucket")

        self.bucket_name = name or "${self:provider.deploymentBucket.name}"

        if serverSideEncryption:
            self.serverSideEncryption = serverSideEncryption

        if kmsKeyID:
            if serverSideEncryption == "aws:kms":
                self.kmsKeyID = kmsKeyID
            else:
                raise InvalidInputException("kmsKeyID can only be set if serverSideEncryption is 'aws:kms'")

        if versioning:
            self.versioning = versioning

        if accelerate:
            self.accelerate = accelerate

        if blockPublicAccess:
            self.blockPublicAccess = blockPublicAccess

        if tags:
            self.tags = list(map(lambda key, value: {"Key": key, "Value": value}, tags.items()))

    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        export["name"] = export["bucket_name"]
        export.pop("bucket_name")

        service.custom.deploymentBucket = dict(self)



import yaml

from serverless.aws.functions.event_bridge import EventBridgeFunction
from serverless.aws.functions.generic import Function
from serverless.aws.functions.http import HTTPFunction
from serverless.aws.functions.kinesis import KinesisFunction
from serverless.aws.functions.s3 import S3Function
from serverless.aws.functions.sqs import SQSFunction
from serverless.aws.iam import ServicePolicyBuilder
from serverless.service.environment import Environment
from serverless.service.types import Provider as BaseProvider


class Runtime(yaml.YAMLObject):
    NODE_10 = "nodejs10"
    NODE_12 = "nodejs12"
    NODE_14 = "nodejs14"

    PYTHON_2_7 = "python2.7"
    PYTHON_3_6 = "python3.6"
    PYTHON_3_7 = "python3.7"
    PYTHON_3_8 = "python3.8"
    PYTHON_3_9 = "python3.9"

    yaml_tag = "!Runtime"

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_str(data)


class FunctionBuilder:
    def __init__(self, service):
        super().__init__()
        self.service = service

    def generic(self, name, description, handler=None, timeout=None, layers=None, **kwargs) -> Function:
        fn = Function(self.service, name, description, handler, timeout, layers, **kwargs)
        self.service.functions.add(fn)

        return fn

    def http(
        self, name, description, path, method, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(self.service, name, description, path, method, authorizer, handler, timeout, layers, **kwargs)
        self.service.functions.add(fn)

        return fn

    def http_post(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.POST, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_get(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.GET, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_put(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.PUT, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_patch(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.PATCH, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_delete(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.DELETE, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_options(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.OPTIONS, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def http_any(
        self, name, description, path, authorizer=None, handler=None, timeout=None, layers=None, **kwargs
    ) -> HTTPFunction:
        fn = HTTPFunction(
            self.service, name, description, path, HTTPFunction.ANY, authorizer, handler, timeout, layers, **kwargs
        )
        self.service.functions.add(fn)

        return fn

    def event_bridge(
        self,
        name,
        description,
        eventBus,
        pattern=None,
        deadLetterQueueArn=None,
        retryPolicy=None,
        handler=None,
        timeout=None,
        layers=None,
        **kwargs,
    ) -> EventBridgeFunction:
        fn = EventBridgeFunction(
            self.service,
            name,
            description,
            eventBus,
            pattern,
            deadLetterQueueArn,
            retryPolicy,
            handler,
            timeout,
            layers,
            **kwargs,
        )
        self.service.functions.add(fn)

        return fn

    def s3(self, name, description, bucket, event, rules=None, handler=None, timeout=None, layers=None, **kwargs):
        fn = S3Function(self.service, name, description, bucket, event, rules, handler, timeout, layers, **kwargs)
        self.service.functions.add(fn)

        return fn

    def kinesis(self, stream, name, description, **kwargs):
        fn = KinesisFunction(stream, self.service, name, description, **kwargs)
        self.service.functions.add(fn)

        return fn

    def sqs(self, name, description, arn, **kwargs):
        fn = SQSFunction(self.service, name, description, arn, **kwargs)
        self.service.functions.add(fn)

        return fn


class Provider(BaseProvider, yaml.YAMLObject):
    yaml_tag = "!Provider"

    def __init__(
        self, runtime=Runtime.PYTHON_3_8, extra_tags=None, timeout=10, stage="${sls:stage}", environment=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.deploymentBucket = None
        self._service = None
        self.function_builder = None

        self.name = "aws"
        self.runtime = runtime
        self.stackName = "${self:service}-${sls:stage}"
        self.timeout = timeout
        self.stage = stage
        self.tags = extra_tags or {}
        self.stackTags = {}
        self.lambdaHashingVersion = 20201221
        self.environment = environment or Environment()
        self.iam = None
        self.apiGateway = dict(shouldStartNameWithService=True)
        self.eventBridge = dict(useCloudFormation=True)

    def configure(self, service):
        self._service = service

        if service.config.domain:
            self.deploymentBucket = dict(name=f"sls-deployments.${{aws:region}}.${{sls:stage}}.{service.config.domain}")

        self.tags["SERVICE"] = "${self:service}"
        self.tags["REGION"] = "${aws:region}"
        self.tags["STAGE"] = "${sls:stage}"
        self.stackTags = self.tags.copy()
        self.iam = ServicePolicyBuilder(self._service)
        self.function_builder = FunctionBuilder(self._service)

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("_service", None)
        data.pop("function_builder", None)
        if not data.get("deploymentBucket"):
            data.pop("deploymentBucket", None)

        return dumper.represent_dict(data)



import hashlib

from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class SQSPublisher(IAMPreset):
    def __init__(self, resource):
        super().__init__(resource)

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=["sqs:SendMessageBatch", "sqs:SendMessage"],
            resources=str(self.resource),
            sid="SQSPublisherI" + hashlib.sha224(str(self.resource).encode("ascii")).hexdigest(),
        )



from troposphere.logs import LogGroup as AWSLogGroup

from serverless.aws.features.encryption import Encryption
from serverless.aws.resources import Resource
from serverless.aws.resources.kms import EncryptableResource
from serverless.service.types import Identifier


class LogGroup(Resource):
    def __init__(self, LogGroupName, **kwargs) -> None:
        kwargs.setdefault("RetentionInDays", 30)
        kwargs.setdefault("title", Identifier(LogGroupName + "LogGroup").resource)

        super().__init__(AWSLogGroup(LogGroupName=LogGroupName, **kwargs))

    def configure(self, service):
        prefix = "/services/" + service.service.spinal
        if not self.resource.LogGroupName.startswith(prefix):
            self.resource.LogGroupName = prefix + "/" + self.resource.LogGroupName.strip("/")

        if service.has(Encryption):
            self.resource.KmsKeyId = EncryptableResource.encryption_arn()
            self.resource.DependsOn = [EncryptableResource.encryption_key_name() + "Alias"]

        super().configure(service)



import yaml

from serverless.service.types import Plugin


class PluginsManager(yaml.YAMLObject):
    yaml_tag = "!Plugins"

    def __init__(self, service):
        super().__init__()
        self._service = service
        self._plugins = []

    def add(self, plugin: Plugin):
        plugin.enable(self._service)
        self._plugins.append(plugin)

    def get(self, plugin):
        for p in self._plugins:
            if isinstance(p, plugin):
                return p

    def has(self, plugin):
        for p in self._plugins:
            if isinstance(p, plugin):
                return True

        return False

    def all(self):
        return self._plugins

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_list([plugin.name for plugin in data.all()])



import io
import os
from collections import OrderedDict
from pathlib import Path
from typing import Optional, Union

import yaml

from serverless.aws.features.stepfunctions import StepFunctions
from serverless.service.configuration import Configuration
from serverless.service.functions import FunctionManager
from serverless.service.package import Package
from serverless.service.plugins import PluginsManager
from serverless.service.resources import ResourceManager
from serverless.service.types import Identifier, Provider, YamlOrderedDict


class Builder:
    def __init__(self, service):
        self.service = service
        self.function = service.provider.function_builder


class PreSetAttributesBuilder(Builder):
    def __init__(self, service, preset):
        super().__init__(service)
        self._preset = preset

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __getattr__(self, item):
        def wrapper(*args, **kwargs):
            return getattr(self.function, item)(*args, **{**kwargs, **self._preset})

        return wrapper


class Service(OrderedDict, yaml.YAMLObject):
    yaml_tag = "!Service"

    def __init__(
        self,
        name: str,
        description: str,
        provider: Provider,
        config: Optional[Configuration] = None,
        custom: Optional[dict] = None,
        **kwds,
    ):
        super().__init__(**kwds)

        self.service = Identifier(name)
        self.package = Package(["!./**/**", f"{self.service.snake}/**"])
        self.variablesResolutionMode = 20210326
        self.custom = YamlOrderedDict(vars="${file(./variables.yml):${sls:stage}}", **(custom or {}))
        self.config = config or Configuration()

        provider.configure(self)
        self.provider = provider

        self.plugins = PluginsManager(self)
        self.functions = FunctionManager(self)
        self.resources = ResourceManager(self, description)

        self.builder = Builder(self)
        self.stepFunctions = StepFunctions(self)
        self.features = []

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, item):
        return self.get(item)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.builder = Builder(self)
        return self

    def preset(self, **kwargs):
        self.builder = PreSetAttributesBuilder(self, kwargs)

        return self.builder

    def enable(self, feature):
        self.features.append(feature)
        feature.enable(self)

    def render(self, output=None, auto_generated_warning=True):
        if "SERVERLESS_BUILDER_DISABLE_RENDER" in os.environ:
            return

        import __main__ as main

        output = output if output else open(Path(main.__file__).stem, "w+")
        content = (
            "# DO NOT edit this file directly, it was generated based on serverless.yml.py\n\n" + str(self)
            if auto_generated_warning
            else str(self)
        )

        output.write(content)

    def __str__(self):
        buf = io.StringIO()
        yaml.dump(self, buf, sort_keys=False, indent=2, width=1000)
        buf.seek(0)
        tmp_buf = io.StringIO()

        for line in buf:
            if line.split(":")[0] in ("provider", "plugins", "package", "custom", "functions", "resources", "vars"):
                tmp_buf.write("\n")

            tmp_buf.write(line)
        tmp_buf.seek(0)

        return tmp_buf.read()

    def has(self, feature):
        return len(list(filter(lambda x: isinstance(x, feature), self.features))) > 0

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("builder", None)
        data.pop("config", None)
        data.pop("function_builder", None)

        if not data.stepFunctions.stateMachines:
            data.pop("stepFunctions", None)

        for plugin in data.plugins.all():
            plugin.pre_render(data)

        for feature in data.features:
            feature.pre_render(data)

        data.pop("features", None)

        return dumper.represent_dict(data)



from serverless.service.types import Integration


class Powertools(Integration):
    def enable(self, service):
        service.provider.environment["envs"]["POWERTOOLS_SERVICE_NAME"] = "${self:service}"
        service.provider.environment["envs"]["POWERTOOLS_LOGGER_LOG_EVENT"] = "${self:custom.vars.log_event}"



from troposphere import Ref, GetAtt


class EncryptableResource:
    @staticmethod
    def encryption_key_name():
        return "ServiceEncryptionKey"

    @staticmethod
    def encryption_key():
        return Ref(EncryptableResource.encryption_key_name())

    @staticmethod
    def encryption_alias():
        return "alias/${self:service}-${sls:stage}"

    @staticmethod
    def encryption_arn():
        return "arn:aws:kms:${aws:region}:${aws:accountId}:alias/${self:service}-${sls:stage}"



from troposphere.kinesis import Stream, StreamEncryption

from serverless.aws.features.encryption import Encryption
from serverless.aws.iam.kinesis import KinesisReader
from serverless.aws.resources import Resource
from serverless.aws.resources.kms import EncryptableResource
from serverless.service import Identifier


class KinesisStream(Resource):
    def __init__(self, Name, **kwargs):
        if "${sls:stage}" not in Name:
            Name = "${self:service}-${sls:stage}-" + Name

        kwargs.setdefault("title", Identifier(Name, safe=True).pascal)

        super().__init__(Stream(Name=Name, **kwargs))

    def configure(self, service):
        super().configure(service)

        if service.has(Encryption) and self.resource.StreamEncryption:
            self.resource.StreamEncryption = StreamEncryption(
                KeyId=EncryptableResource.encryption_key(), EncryptionType="KMS"
            )

    def permissions(self):
        return [KinesisReader(self.resource.Name)]

    @property
    def Name(self):
        return self.resource.Name

    def arn(self):
        return "arn:aws:kinesis:${aws:region}:${aws:accountId}:stream/" + self.resource.Name



from typing import Optional

from troposphere.cloudwatch import Alarm, MetricDimension
from troposphere.dynamodb import (
    AttributeDefinition,
    KeySchema,
    TimeToLiveSpecification,
)

from serverless.aws.features.encryption import Encryption
from serverless.aws.iam import FunctionPolicyBuilder
from serverless.aws.iam.dynamodb import DynamoDBFullAccess
from serverless.aws.resources import DummyResource
from serverless.aws.resources.dynamodb import Table
from serverless.aws.resources.kms import EncryptableResource
from serverless.aws.resources.sqs import Queue
from serverless.aws.types import SQSArn
from serverless.service.environment import Environment
from serverless.service.plugins.iam_roles import IAMRoles
from serverless.service.plugins.lambda_dlq import LambdaDLQ
from serverless.service.plugins.python_requirements import PythonRequirements
from serverless.service.types import Identifier, YamlOrderedDict


class ScheduleEvent(YamlOrderedDict):
    yaml_tag = "schedule"

    def __init__(self, expression):
        super().__init__()
        self.rate = expression


class Function(YamlOrderedDict):
    yaml_tag = "!Function"

    def __init__(
        self,
        service,
        name,
        description,
        handler=None,
        timeout=None,
        layers=None,
        force_name=None,
        idempotency=None,
        use_dlq=False,
        use_async_dlq=False,
        **kwargs,
    ):
        super().__init__()
        self._service = service

        self.key = Identifier(name)
        self.name = (
            force_name
            if force_name
            else Identifier(
                self._service.service.spinal.lower() + "-${sls:stage}" + "-" + self.key.spinal.lower()
            )
        )
        self.description = description

        if not handler:
            handler = f"{self._service.service.snake}.{self.key.spinal}.handler"

        self.handler = handler
        self.events = []

        if self._service.plugins.get(IAMRoles):
            self.iamRoleStatements = FunctionPolicyBuilder(self.name, self._service)
            self.iamRoleStatementsName = self.iamRoleStatements.role
            self.iamRoleStatementsInherit = True

        configured = list(filter(lambda x: x.get("Ref") == "PythonRequirementsLambdaLayer", layers or []))
        if (
            self._service.plugins.get(PythonRequirements)
            and self._service.plugins.get(PythonRequirements).layer
            and not configured
        ):
            if not layers:
                layers = []

            layers.append({"Ref": "PythonRequirementsLambdaLayer"})

        if layers:
            self.layers = layers

        if timeout:
            self.timeout = timeout

        if idempotency:
            self.with_idempotency(idempotency)

        for name, value in kwargs.items():
            setattr(self, name, value)

        self.dlq = None

        if use_dlq:
            self.use_dlq()

        if use_async_dlq:
            self.use_async_dlq()

        log_group = dict(Type="AWS::Logs::LogGroup", Properties=dict(RetentionInDays=30))
        if service.has(Encryption):
            log_group["Properties"]["KmsKeyId"] = EncryptableResource.encryption_arn()
            log_group["DependsOn"] = [EncryptableResource.encryption_key_name() + "Alias"]

        service.resources.add(DummyResource(title=self.log_group_name(), **log_group))

    @property
    def iam(self):
        if not self._service.plugins.get(IAMRoles):
            self._service.plugins.add(IAMRoles())

        if not self.iamRoleStatements:
            self.iamRoleStatements = FunctionPolicyBuilder(self.name, self._service)
            self.iamRoleStatementsName = self.iamRoleStatements.role
            self.iamRoleStatementsInherit = True

        return self.iamRoleStatements

    def get_attr(self, attr):
        return {"Fn::GetAtt": [self.resource_name(), attr]}

    def arn(self):
        return self.get_attr("Arn")

    def log_group_name(self):
        return f"{self.key.pascal}LogGroup"

    def iam_role_name(self):
        return f"{self.key.pascal}IamRoleLambdaExecution"

    def resource_name(self):
        return f"{self.key.pascal}LambdaFunction"

    def trigger(self, event):
        self.events.append(event)

        return event

    def apply(self, **kwargs):
        for event in self.events:
            for k, v in kwargs.items():
                event[k] = v

    def use_dlq(self, onErrorDLQArn: Optional[str] = None, MessageRetentionPeriod: int = 1209600) -> None:
        """
        @param onErrorDLQArn: Optional[str]
        @param MessageRetentionPeriod: integer – defaults to 14 days in seconds
        @return None
        """

        if not self._service.plugins.has(LambdaDLQ):
            self._service.plugins.add(LambdaDLQ())

        if not onErrorDLQArn:
            onErrorDLQArn = self._ensure_dlq(MessageRetentionPeriod)

        self.deadLetter = dict(targetArn=self._ensure_dlq(MessageRetentionPeriod).get("arn"))

        return self

    def use_async_dlq(self, onFailuredlqArn: Optional[str] = None, MessageRetentionPeriod: int = 1209600) -> None:
        """
        @param onFailuredlqArn: Optional[str]
        @param MessageRetentionPeriod: integer – defaults to 14 days in seconds
        @return None
        """
        if not onFailuredlqArn:
            onFailuredlqArn = self._ensure_dlq(MessageRetentionPeriod).get("arn")

        self.destinations = dict(onFailure=onFailuredlqArn)

        return self

    def _ensure_dlq(self, MessageRetentionPeriod):
        name = f"{self.name.spinal}-dlq"
        resource = f"{self.name.resource}DLQ"
        if self.dlq:
            return {"Ref": resource, "arn": SQSArn(name)}

        self.dlq = Queue(
            title=resource,
            QueueName=name,
            MessageRetentionPeriod=MessageRetentionPeriod,
        )
        self._service.resources.add(self.dlq)

        self.iam.allow(
            sid=f"{self.name.resource}DLQWriter",
            permissions=["sqs:GetQueueUrl", "sqs:SendMessageBatch", "sqs:SendMessage"],
            resources=[self.dlq.arn()],
        )

        self._service.resources.add(
            Alarm(
                f"{self.name.resource}DLQAlarm",
                AlarmDescription=f"Lambda: {self.name} rejected event in DLQ",
                AlarmActions=["arn:aws:sns:${aws:region}:${aws:accountId}:foxglove-${sls:stage}-cloudwatch-alerts"],
                Namespace="AWS/SQS",
                MetricName="ApproximateNumberOfMessagesVisible",
                Dimensions=[MetricDimension(Name="QueueName", Value=name)],
                Statistic="Sum",
                Period=60,
                EvaluationPeriods=1,
                Threshold=1,
                ComparisonOperator="GreaterThanOrEqualToThreshold",
            )
        )

        self._service.resources.add(
            DummyResource(
                self.resource_name(),
                Type="AWS::Lambda::Function",
                DependsOn=[
                    self.iam_role_name(),
                    self.log_group_name(),
                    resource,
                ],
            )
        )

        return {"Ref": f"{self.name.pascal}DLQ", "arn": SQSArn(name)}

    def with_vpc(self, security_group_names=None, subnet_names=None):
        if security_group_names:
            self.vpcDiscovery = dict(securityGroups=[dict(tagKey="Name", tagValues=security_group_names.copy())])

        return self

    def with_idempotency(self, table_name=None):
        table_name = table_name or f"{self.name.pascal}Idempotency"

        idempotency_table = Table(
            TableName=table_name,
            DeletionPolicy="Retain",
            BillingMode="PAY_PER_REQUEST",
            AttributeDefinitions=[
                AttributeDefinition(AttributeName="id", AttributeType="S"),
            ],
            KeySchema=[
                KeySchema(AttributeName="id", KeyType="HASH"),
            ],
            TimeToLiveSpecification=TimeToLiveSpecification(AttributeName="expiration", Enabled=True),
        ).with_full_access()
        self._service.resources.add(idempotency_table)
        env = self.get("environment", Environment())
        env.envs["IDEMPOTENCY_TABLE"] = idempotency_table.table_arn
        self.environment = env

        return self

    @classmethod
    def to_yaml(cls, dumper, data):
        events = data.events
        data.pop("_service", None)
        data.pop("key", None)
        data.pop("dlq", None)

        if not data.events:
            del data["events"]
        else:
            data.events = [{event.yaml_tag: event} for event in events]

        return dumper.represent_dict(data)



from typing import Optional

from serverless.service.types import SmartString


class Configuration:
    def __init__(self, domain: Optional[str] = None):
        self.domain = SmartString(domain) if domain else None



from serverless.aws.iam.xray import XRayService
from serverless.service.types import Feature


class XRay(Feature):
    def __init__(self, trace_lambda=True):
        super().__init__()
        self.trace_lambda = trace_lambda

    def enable(self, service):
        service.provider.tracing = {"lambda": self.trace_lambda}
        service.provider.iam.apply(XRayService())



from troposphere import GetAtt
from troposphere.cloudwatch import Alarm, MetricDimension

from serverless.aws.iam.event_bridge import Publish
from serverless.aws.resources.sqs import Queue
from serverless.service.types import Integration


class EventDispatcher(Integration):
    def __init__(self, event_bus) -> None:
        super().__init__()
        self.event_bus = event_bus

    def enable(self, service):
        queue = Queue(
            title="EventDispatcherDLQ",
            QueueName="event-dispatcher-dlq",
            MessageRetentionPeriod=1209600,  # 14 days in seconds
        )

        service.resources.add(queue)

        service.resources.add(
            Alarm(
                "EventDispatcherDLQCloudWatchAlarm",
                AlarmDescription="Unable to send event from handler.",
                AlarmActions=["arn:aws:sns:${aws:region}:${aws:accountId}:foxglove-${sls:stage}-cloudwatch-alerts"],
                Namespace="AWS/SQS",
                MetricName="ApproximateNumberOfMessagesVisible",
                Dimensions=[MetricDimension(Name="QueueName", Value=GetAtt("EventDispatcherDLQ", "QueueName"))],
                Statistic="Sum",
                Period=60,
                EvaluationPeriods=1,
                Threshold=1,
                ComparisonOperator="GreaterThanOrEqualToThreshold",
            )
        )
        service.provider.iam.allow(
            sid="EventDispatcherDLQBatch",
            permissions=["sqs:SendMessageBatch", "sqs:GetQueueUrl"],
            resources=[queue.arn()],
        )

        service.provider.environment.envs["EVENTS_DLQ"] = queue.resource.QueueName
        service.provider.iam.apply(Publish(self.event_bus))


class DLQ(EventDispatcher):
    pass



from typing import Union

import yaml
from troposphere import AWSObject

from serverless.aws.resources import Resource


class ResourceManager(yaml.YAMLObject):
    yaml_tag = "!Resources"

    def __init__(self, service, description):
        super().__init__()
        self.description = description
        self._service = service
        self.resources = []

    def add(self, resource: Union[AWSObject, Resource]):
        if isinstance(resource, Resource):
            resource.configure(self._service)
            self.add_all(resource.resources())
            self._service.provider.environment.envs.update(resource.variables())

            for preset in resource.permissions():
                self._service.provider.iam.apply(preset)
        else:
            self.resources.append(resource)

        return resource

    def add_all(self, resources):
        for resource in resources:
            self.add(resource)

    def all(self):
        return self.resources

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict(
            dict(
                Description=data.description,
                Resources={resource.title: resource.to_dict() for resource in data.resources},
            )
        )



from serverless.service.plugins.generic import Generic


class Prune(Generic):
    yaml_tag = "!PrunePlugin"

    def __init__(self, automatic=True, number=3):
        super().__init__("serverless-prune-plugin")
        self.automatic = automatic
        self.number = number

    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        service.custom.prune = export



from serverless.aws.resources.logs import LogGroup
from serverless.service.types import YamlOrderedDict, Identifier


class Stage(YamlOrderedDict):
    yaml_tag = "!Stage"

    def __init__(self, type, function=None):
        self.Type = type
        self._function = function

    def next(self, stage):
        self.Next = stage.id

    @property
    def id(self):
        return self._function.key.pascal

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("_function", None)
        return super().to_yaml(dumper, data)


class Fallback(Stage):
    yaml_tag = "!Fallback"

    def __init__(self, name, result):
        super().__init__("Pass")
        self.Result = result
        self.End = True
        self._name = name

    @property
    def id(self):
        return self._name

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("_name")
        return super().to_yaml(dumper, data)


class Task(Stage):
    yaml_tag = "!Task"

    def __init__(self, function=None, end=None, name=None, parameters=None, resource=None):
        if not any([function, resource]):
            raise Exception("You need to provide either function or resource parameter")

        super().__init__("Task", function)

        if name:
            self.name = name

        if parameters:
            self.Parameters = parameters

        self.Resource = function.arn() if function else resource
        if end is not None:
            self.End = end

    @property
    def id(self):
        return self.name if self.name else self._function.key.pascal

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("name", None)
        return super().to_yaml(dumper, data)


class Iterator(YamlOrderedDict):
    yaml_tag = "!Iterator"

    def __init__(self, map_name, steps, auto_catch=True, auto_fallback=False):
        self.States = {step.id: step for step in steps}
        self.map_name = map_name
        self.auto_catch = auto_catch
        self.auto_fallback = auto_fallback

    @property
    def id(self):
        return self.map_name

    @classmethod
    def to_yaml(cls, dumper, data):
        data["StartAt"] = str(list(data.States.keys())[0])
        data.move_to_end("StartAt", last=False)

        fallback = None
        for step in data.States.values():
            if step.get("Type") in ("Pass", "Fail"):
                fallback = step
                break

        if data.auto_fallback and not fallback:
            fallback = Fallback(f"{data.map_name}Fallback", f"Failed processing map: {data.map_name}")
            data["States"][fallback.id] = fallback

        if data.auto_catch:
            for step in data.States.values():
                if step.get("Catch") or step.get("Type") in ("Pass", "Fail", "Map"):
                    continue

                step["Catch"] = [{"ErrorEquals": ["States.ALL"], "Next": fallback.id}]

        data.pop("map_name", None)
        data.pop("auto_catch", None)
        data.pop("auto_fallback", None)

        return super().to_yaml(dumper, data)


class Map(Stage):
    yaml_tag = "!Map"

    def __init__(
        self, name, steps, items_path="$.items", input_path="$.body", result_path="$.results", concurrency=20, end=False
    ):
        super().__init__("Map")
        self.name = name

        if input_path:
            self.InputPath = input_path

        if items_path:
            self.ItemsPath = items_path

        if result_path:
            self.ResultPath = result_path

        self.MaxConcurrency = concurrency
        self.End = end
        self.Iterator = steps if isinstance(steps, Iterator) else Iterator(name, steps)

    @property
    def id(self):
        return self.name

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("name", None)
        return super().to_yaml(dumper, data)


class State(YamlOrderedDict):
    yaml_tag = "!State"

    def __init__(self, name) -> None:
        super().__init__()
        self.name = name

    @property
    def id(self):
        return self.name

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("name", None)
        return super().to_yaml(dumper, data)


class Choice(State):
    yaml_tag = "!Choice"

    def __init__(self, name, default, choices=None) -> None:
        super().__init__(name)
        self.Type = "Choice"
        self.Default = default
        self.Choices = choices or []


class Succeed(State):
    yaml_tag = "!Succeed"

    def __init__(self, name) -> None:
        super().__init__(name)
        self.Type = "Succeed"


class Branch(YamlOrderedDict):
    yaml_tag = "!Branch"

    def __init__(self, *states) -> None:
        super().__init__()
        self.States = list(states)

    @classmethod
    def to_yaml(cls, dumper, data):
        export = dict(StartAt=data.States[0].id, States={i.name: i for i in data.States})
        return super().to_yaml(dumper, export)


class Parallel(Stage):
    yaml_tag = "!Parallel"

    def __init__(self, name, branches, end=False):
        super().__init__("Parallel")
        self.name = name
        self.Branches = branches
        self.End = end

    @property
    def id(self):
        return self.name

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("name", None)
        return super().to_yaml(dumper, data)


class Definition(YamlOrderedDict):
    yaml_tag = "!Definition"

    def __init__(self, description, auto_fallback=True, auto_catch=True):
        self.Comment = description
        self.States = YamlOrderedDict()
        self.auto_fallback = auto_fallback
        self.auto_catch = auto_catch

    def add(self, state):
        self.States[state.id] = state

        return state

    @classmethod
    def to_yaml(cls, dumper, data):
        if data.States:
            data["StartAt"] = list(data.States.keys())[0]
            data.move_to_end("StartAt", last=False)

        fallback = None
        for step in data.States.values():
            if step.get("Type") in ("Pass", "Fail"):
                fallback = step
                break

        if data.auto_fallback and not fallback:
            fallback = Fallback("StateMachineErrorFallback", "Error in state machine")
            data.add(fallback)

        if data.auto_catch:
            for step in data.States.values():
                if step.get("Catch") or step.get("Type") in ("Pass", "Fail", "Map"):
                    continue

                step["Catch"] = [{"ErrorEquals": ["States.ALL"], "Next": fallback.id}]

        data.pop("auto_fallback", None)
        data.pop("auto_catch", None)

        return super().to_yaml(dumper, data)


class StateMachine(YamlOrderedDict):
    yaml_tag = "!YamlOrderedDict"

    def __init__(self, name, description, events=None, type=None, auto_fallback=True, auto_catch=True, service=None):
        self.name = name
        self.tracingConfig = dict(enabled=True)

        if type:
            self.type = type

        logs = LogGroup(LogGroupName=f"/stepmachine/{Identifier(self.name).spinal}")
        service.resources.add(logs)

        self.loggingConfig = dict(level="ERROR", includeExecutionData=True, destinations=[logs.get_att("Arn")])
        self.definition = Definition(description, auto_fallback, auto_catch)
        self.events = events or []

    def task(self, function):
        return self.definition.add(Task(function))

    def map(self, name, steps, **kwargs):
        return self.definition.add(Map(name, steps, **kwargs))

    def parallel(self, name, branches, end):
        return self.definition.add(Parallel(name=name, branches=branches, end=end))

    def event(self, event):
        self.events.append(event)

    def arn(self):
        return "arn:aws:states:${aws:region}:${aws:accountId}:stateMachine:" + self.name


class StepFunctions(YamlOrderedDict):
    yaml_tag = "!StepFunctions"

    def __init__(self, service):
        self.service = service
        self.validate = True
        self.stateMachines = YamlOrderedDict()

    def machine(self, name, description, type=None, auto_fallback=True, auto_catch=True):
        if name in self.stateMachines:
            return self.stateMachines.get(name)

        machine = StateMachine(
            f"{self.service.service}-${{sls:stage}}-{name}",
            description,
            auto_fallback=auto_fallback,
            auto_catch=auto_catch,
            type=type,
            service=self.service,
        )
        self.stateMachines[name] = machine

        return machine

    @classmethod
    def to_yaml(cls, dumper, data):
        data.pop("service", None)
        return super().to_yaml(dumper, data)


class Scheduled(YamlOrderedDict):
    yaml_tag = "!YamlOrderedDict"

    def __init__(self, expression, inputPath=None):
        if not isinstance(expression, dict):
            expression = dict(rate=f"cron({expression})")

        self.schedule = expression

        if inputPath:
            self.inputPath = inputPath



from troposphere.s3 import (
    Bucket,
    Private,
    BucketEncryption,
    ServerSideEncryptionRule,
    ServerSideEncryptionByDefault,
    PublicAccessBlockConfiguration,
    VersioningConfiguration,
)

from serverless.aws.features.encryption import Encryption
from serverless.aws.resources import Resource

from serverless.aws.resources.kms import EncryptableResource
from serverless.service import Identifier


class S3Bucket(Resource):
    def __init__(self, BucketName="${self:service}", domain=None, **kwargs):
        kwargs.setdefault("AccessControl", "Private")

        kwargs.setdefault(
            "PublicAccessBlockConfiguration",
            PublicAccessBlockConfiguration(
                BlockPublicAcls=True,
                BlockPublicPolicy=True,
                IgnorePublicAcls=True,
                RestrictPublicBuckets=True,
            ),
        )

        kwargs.setdefault("VersioningConfiguration", VersioningConfiguration(Status="Enabled"))

        bucket = Bucket(title=Identifier(BucketName, safe=True).pascal, **kwargs)

        bucket.properties.__setitem__(
            "BucketName", BucketName + ".${aws:region}." + (domain or "${ssm:/global/primary-domain}")
        )

        super().__init__(bucket)

    def configure(self, service):
        super().configure(service)

        if service.has(Encryption):
            self.resource.BucketEncryption = (
                BucketEncryption(
                    ServerSideEncryptionConfiguration=[
                        ServerSideEncryptionRule(
                            BucketKeyEnabled=True,
                            ServerSideEncryptionByDefault=ServerSideEncryptionByDefault(
                                KMSMasterKeyID=EncryptableResource.encryption_key(), SSEAlgorithm="aws:kms"
                            ),
                        )
                    ]
                ),
            )



import re

from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class Execute(IAMPreset):
    def __init__(self, api, endpoint):
        self.api = api
        self.endpoint = endpoint

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        if not str(self.endpoint).startswith("arn:aws:execute-api:"):
            arn = "arn:aws:execute-api:${aws:region}:${aws:accountId}:" + self.api + "/${sls:stage}/" + self.endpoint
        else:
            arn = self.endpoint

        policy_builder.allow(
            permissions=["execute-api:Invoke"],
            resources=[arn],
            sid=sid or "Invoke" + Identifier(re.sub(r"\W", "", self.endpoint)).pascal,
        )



import abc
import hashlib
import json
from abc import ABC

from serverless.aws.resources import Resource
from serverless.service.types import YamlOrderedDict, ResourceName


class PolicyBuilder(YamlOrderedDict):
    yaml_tag = "!IAM"

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.statements = []

    def append(self, policy):
        if len(list(filter(lambda x: x["Sid"] == policy["Sid"], self.statements))):
            return

        self.statements.append(policy)

    def allow(self, permissions, resources, sid=None):
        sid = sid or "Policy" + str(hashlib.sha224(json.dumps([permissions, resources]).encode("ascii")).hexdigest())
        self.append(dict(Sid=sid, Effect="Allow", Action=permissions, Resource=resources))

    def deny(self, permissions, resources, sid=None):
        sid = sid or "Policy" + str(hashlib.sha224(json.dumps([permissions, resources]).encode("ascii")).hexdigest())
        self.append(dict(Sid=sid, Effect="Deny", Action=permissions, Resource=resources))

    def apply(self, preset: "IAMPreset"):
        preset.apply(self)

    def allow_read(self, resource: Resource):
        resource.enable_read(self)

    def allow_write(self, resource: Resource):
        resource.enable_write(self)

    def allow_delete(self, resource: Resource):
        resource.enable_delete(self)

    @property
    def role_arn(self):
        return "arn:aws:iam::${aws:accountId}:role/" + self.role

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_list(data.statements)


class ServicePolicyBuilder(PolicyBuilder):
    yaml_tag = "!IAM"

    def __init__(self, service, name=None, **kwds):
        super().__init__(**kwds)
        self.service = service
        self.name = name or "${self:service}-${sls:stage}-${aws:region}-service"

    @property
    def role(self):
        return str(ResourceName(self.name, self.service))

    @classmethod
    def to_yaml(cls, dumper, data):
        export = dict(data.items())
        export.pop("service", None)

        return dumper.represent_dict(dict(role=export))


class FunctionPolicyBuilder(PolicyBuilder):
    yaml_tag = "!IAM"

    def __init__(self, function_name, service, **kwds):
        super().__init__(**kwds)
        self.function_name = function_name
        self.service = service

    @property
    def role(self):
        name = str(self.function_name)
        name = name.replace("${sls:stage}", "${sls:stage}-${aws:region}")
        return str(ResourceName(f"{name}-lambda", self.service))


class IAMPreset(ABC):
    def __init__(self, resource):
        super().__init__()
        self.resource = resource

    @abc.abstractmethod
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        pass



from troposphere.sqs import Queue as SQSQueue

from serverless.aws.features.encryption import Encryption
from serverless.aws.resources import Resource
from serverless.aws.resources.kms import EncryptableResource
from serverless.service.types import Identifier


class Queue(Resource):
    def __init__(self, QueueName, **kwargs):
        if "${sls:stage}" not in QueueName:
            QueueName = "${self:service}-${sls:stage}-" + QueueName

        kwargs.setdefault("title", Identifier(QueueName).resource)

        super().__init__(SQSQueue(QueueName=QueueName, **kwargs))

    def configure(self, service):
        super().configure(service)

        if service.has(Encryption):
            self.resource.KmsMasterKeyId = EncryptableResource.encryption_key()

    def permissions(self):
        return []

    def arn(self):
        return "arn:aws:sqs:${aws:region}:${aws:accountId}:" + self.resource.QueueName



import yaml


class FunctionManager(yaml.YAMLObject):
    yaml_tag = "!Functions"

    def __init__(self, service):
        super().__init__()
        self.service = service
        self.functions = []

    def add(self, function):
        self.functions.append(function)

    def all(self):
        return self.functions

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict({function.key.pascal: function for function in data.functions})



from troposphere import Ref
from troposphere.apigateway import GatewayResponse

from serverless.service.types import Feature


class DefaultFourHundredResponse(Feature):
    def enable(self, service):
        service.resources.add(
            GatewayResponse(
                "GatewayResponseDefault4XX",
                ResponseParameters={
                    "gatewayresponse.header.Access-Control-Allow-Origin": "'*'",
                    "gatewayresponse.header.Access-Control-Allow-Headers": "'*'",
                },
                ResponseType="DEFAULT_4XX",
                RestApiId=Ref("ApiGatewayRestApi"),
            )
        )



from serverless.service.plugins.generic import Generic


class DomainManager(Generic):
    """
    Plugin homepage: https://github.com/amplify-education/serverless-domain-manager
    """

    yaml_tag = "!DomainManagerPlugin"

    def __init__(self, domain, api="rest", stage="${sls:stage}", base_path=None, create_route_53_record=False):
        super().__init__("serverless-domain-manager")
        self[api] = dict(
            domainName=f"{api}.{domain}", stage=stage, basePath=base_path, createRoute53Record=create_route_53_record
        )

    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        service.custom.customDomain = export



from serverless.service.types import YamlOrderedDict


class Generic(YamlOrderedDict):
    yaml_tag = "!GenericPlugin"

    def __init__(self, name, other=(), /, **kwds):
        super().__init__(other, **kwds)
        self.name = name

    def enable(self, service):
        pass

    def pre_render(self, service):
        pass

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_str(data.name)



from serverless.service import YamlOrderedDict


class CognitoAuth(YamlOrderedDict):
    yaml_tag = "!CognitoAuth"

    def __init__(self, arn=None, pool=None):
        self.name = "Cognito"
        if arn:
            self.arn = arn
        if pool:
            self.arn = f"arn:aws:cognito-idp:${{aws:region}}:${{aws:accountId}}:userpool/{pool}"

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict(data)



import itertools

from serverless.aws.functions.generic import Function
from serverless.service.types import YamlOrderedDict


class HTTPAuthorizer(YamlOrderedDict):
    yaml_tag = "!Authorizer"

    def __init__(self, name, identitySource, type):
        super().__init__()
        self.identitySource = identitySource
        self.name = name
        self.type = type


class HTTPEvent(YamlOrderedDict):
    yaml_tag = "http"

    def __init__(self, path, method, authorizer=None, request_parameters_querystrings=None):
        super().__init__()
        self.path = path
        self.method = method

        if authorizer:
            self.authorizer = authorizer

        if request_parameters_querystrings:
            self.request = {"parameters": {"querystrings": request_parameters_querystrings}}


class HTTPFunction(Function):
    yaml_tag = "!HTTPFunction"

    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"
    ANY = "ANY"

    def __init__(
        self,
        service,
        name,
        description,
        path,
        method,
        authorizer=None,
        handler=None,
        timeout=None,
        layers=None,
        request_parameters_querystrings=None,
        **kwargs
    ):
        super().__init__(service, name, description, handler, timeout, layers, **kwargs)

        if isinstance(path, str):
            path = [path]
        if isinstance(method, str):
            method = [method]

        # Set up an HTTP event on all combinations of path and method
        for p, m in itertools.product(path, method):
            self.trigger(HTTPEvent(p, m, authorizer, request_parameters_querystrings))



import abc
import re
from collections import OrderedDict

import inflection
import yaml


class YamlOrderedDict(OrderedDict, yaml.YAMLObject):
    yaml_tag = "!YamlOrderedDict"

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, item):
        return self.get(item)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict(data)


class SmartString:
    def __init__(self, val, prefix="", suffix=""):
        self.val = val
        self.prefix = prefix
        self.suffix = suffix

    def __str__(self):
        if not self.val:
            return ""

        return f"{self.prefix}{self.val}{self.suffix}"

    def __call__(self, prefix="", suffix=""):
        return SmartString(self.val, prefix=prefix, suffix=suffix)


class Identifier(yaml.YAMLObject):
    yaml_tag = "Identifier"

    def __init__(self, identifier, safe=False):
        super().__init__()
        if safe:
            identifier = re.sub(r"\${.*?}-?", "", identifier)
            identifier = re.sub(r"\W", "-", identifier)

        self.identifier = identifier

    @property
    def camel(self):
        return inflection.camelize(inflection.underscore(self.identifier))

    @property
    def pascal(self):
        return inflection.camelize(inflection.underscore(self.identifier), uppercase_first_letter=True)

    @property
    def snake(self):
        return inflection.underscore(self.identifier).lower()

    @property
    def spinal(self):
        return inflection.dasherize(self.identifier).lower()

    @property
    def resource(self):
        identifier = re.sub(r"\${.*?}-?", "", self.identifier)
        identifier = re.sub(r"\W", "-", identifier)
        identifier = identifier.strip("-")
        return inflection.camelize(inflection.underscore(identifier), uppercase_first_letter=True)

    def __str__(self):
        return self.identifier

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_str(str(data.identifier))


class ResourceName:
    def __init__(self, name: str, service):
        self.name = name
        self.service = service

    def __str__(self):
        safe = self.name.replace("${aws:region}", "us-east-1")
        safe = safe.replace("${self:service}", self.service.service.spinal)
        safe = safe.replace("${sls:stage}", "staging")

        if len(safe) > 64:
            parts = []
            for part in self.name.split("-"):
                if "$" in part or part == "lambda":
                    parts.append(part)
                else:
                    parts.append(part[0:3])
            return "-".join(parts)
        else:
            return self.name


class ProviderMetadata(type(YamlOrderedDict), type(abc.ABC)):
    pass


class Provider(YamlOrderedDict, abc.ABC, metaclass=ProviderMetadata):
    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, item):
        return self.get(item)

    @abc.abstractmethod
    def configure(self, service):
        pass


class Feature(abc.ABC):
    @abc.abstractmethod
    def enable(self, service):
        pass

    def pre_render(self, service):
        pass


class Integration(abc.ABC):
    @abc.abstractmethod
    def enable(self, service):
        pass

    def pre_render(self, service):
        pass


class PluginMetadata(type(YamlOrderedDict), type(abc.ABC)):
    pass


class Plugin(YamlOrderedDict, abc.ABC, metaclass=PluginMetadata):
    @abc.abstractmethod
    def enable(self, service):
        pass



from serverless.aws.functions.generic import Function
from serverless.service.types import YamlOrderedDict


class SQSEvent(YamlOrderedDict):
    yaml_tag = "sqs"

    def __init__(self, arn, batchSize=10, maximumBatchingWindow=10, enabled=True, filterPatterns=None):
        super().__init__()
        self.arn = arn
        self.batchSize = batchSize
        self.maximumBatchingWindow = maximumBatchingWindow
        self.enabled = enabled
        if filterPatterns:
            self.filterPatterns = filterPatterns


class SQSFunction(Function):
    yaml_tag = "!SQSFunction"

    def __init__(
        self,
        service,
        name,
        description,
        arn,
        batchSize=10,
        maximumBatchingWindow=10,
        enabled=True,
        filterPatterns=None,
        handler=None,
        timeout=None,
        layers=None,
        use_dlq=True,
        use_async_dlq=True,
        **kwargs
    ):
        super().__init__(
            service, name, description, handler, timeout, layers, use_dlq=use_dlq, use_async_dlq=use_async_dlq, **kwargs
        )
        self.trigger(SQSEvent(arn, batchSize, maximumBatchingWindow, enabled, filterPatterns))



from os.path import dirname, basename

from serverless.service.plugins.generic import Generic


class Scriptable(Generic):
    yaml_tag = "!ScriptablePlugin"

    def __init__(self, hooks=None, inject_to_package=None):
        super().__init__("serverless-scriptable-plugin")
        self.inject_to_package = inject_to_package
        self.hooks = hooks or {}

    def enable(self, service):
        service.custom["scriptHooks"] = self.hooks

        if self.inject_to_package:
            service.custom["scriptHooks"][
                "before:package:createDeploymentArtifacts"
            ] = f"cp -a {self.inject_to_package} ."
            service.custom["scriptHooks"][
                "after:package:createDeploymentArtifacts"
            ] = f"rm -rf ./{basename(self.inject_to_package)}"

            service.package.patterns.append(f"{basename(self.inject_to_package)}/**")



from serverless.service.plugins.generic import Generic


class LambdaDLQ(Generic):
    yaml_tag = "!LambdaDLQ"

    def __init__(self):
        super().__init__("serverless-plugin-lambda-dead-letter")

    def enable(self, service):
        pass



class Resource:
    def __init__(self, resource=None) -> None:
        super().__init__()
        self.resource = resource

    def configure(self, service):
        pass

    def resources(self):
        if not self.resource:
            return []

        return [self.resource]

    def permissions(self):
        return []

    def variables(self):
        return {}

    def enable_read(self, builder: "PolicyBuilder"):
        raise NotImplemented()

    def enable_write(self, builder: "PolicyBuilder"):
        raise NotImplemented()

    def enable_delete(self, builder: "PolicyBuilder"):
        raise NotImplemented()

    def get_att(self, name, as_dict=True):
        attr = self.resource.get_att(name)

        if not as_dict:
            return attr

        return attr.to_dict()


class DummyResource:
    def __init__(self, title, **kwargs):
        self.title = title
        self.kwargs = kwargs

    def to_dict(self):
        return self.kwargs



from serverless.service.types import Integration


class Sentry(Integration):
    def __init__(self, sentry_dsn):
        super().__init__()
        self.sentry_dsn = sentry_dsn

    def enable(self, service):
        service.provider.environment["envs"]["SENTRY_DSN"] = self.sentry_dsn
        service.provider.environment["envs"]["SENTRY_ENVIRONMENT"] = "${sls:stage}"
        service.provider.environment["envs"]["SENTRY_RELEASE"] = "${env:VERSION, ''}"



from troposphere.iam import Role as IAMRole, Policy

from serverless.aws.iam import PolicyBuilder
from serverless.aws.resources import Resource
from serverless.service import Identifier


class Role(Resource):
    def __init__(self, RoleName, **kwargs):
        self.role = IAMRole(
            title=Identifier(RoleName, safe=True).pascal,
            RoleName=RoleName,
            Policies=[Policy(PolicyName=RoleName + "-inline", PolicyDocument=dict())],
            validation=False,
            **kwargs
        )

        if "${self:service}" not in RoleName:
            self.role.properties.__setitem__("RoleName", "${self:service}-${sls:stage}-${aws:region}" + RoleName)
        self.policy = PolicyBuilder()

    def resources(self):
        self.role.Policies[0].PolicyDocument = dict(Statement=self.policy.statements)
        return super().resources() + [self.role]



from serverless.service.plugins.generic import Generic


class AWSCodeSign(Generic):
    yaml_tag = "!AWSSignPlugin"

    def __init__(
        self,
        profile,
        sign_policy="Enforce",
        source_bucket=None,
        source_key=None,
        destination_bucket=None,
        destination_prefix=None,
        retain=False,
    ):
        super().__init__("serverless-aws-signer")
        self.profile = profile
        self.sign_policy = sign_policy
        self.source_bucket = source_bucket
        self.source_key = source_key
        self.destination_bucket = destination_bucket
        self.destination_prefix = destination_prefix or "signed-"
        self.retain = retain

    def enable(self, service):
        service.custom.signer = dict(
            retain=self.retain,
            source=dict(
                s3=dict(
                    bucketName=self.source_bucket or service.provider.deploymentBucket.get("name"),
                    key=self.source_key or f"{service.service.spinal}.zip",
                )
            ),
            destination=dict(
                s3=dict(
                    bucketName=self.destination_bucket or service.provider.deploymentBucket.get("name"),
                    prefix=self.destination_prefix,
                )
            ),
            profileName=self.profile,
            signingPolicy=self.sign_policy,
        )



import yaml


class Package(yaml.YAMLObject):
    yaml_tag = "!Package"

    def __init__(self, patterns):
        super().__init__()
        self.patterns = patterns

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict({"patterns": data.patterns})



from serverless.service.types import Feature


class ApiKey:
    def __init__(self, name=None, value=None, description=None):
        if not any([name, value]):
            raise KeyError("You need to provide either name or value for key")

        self.name = name
        self.value = value
        self.description = description

    def to_dict(self):
        return {k: getattr(self, k) for k in ("name", "value", "description") if getattr(self, k)}


class ApiKeys(Feature):
    def __init__(self, keys=None, source="HEADER") -> None:
        super().__init__()
        self.keys = [k if isinstance(k, ApiKey) else ApiKey(k) for k in keys]
        self.source = source

    def enable(self, service):
        api_gw = service.provider.get("apiGateway", {})
        api_gw["apiKeySourceType"] = self.source
        api_gw["apiKeys"] = list(map(lambda x: x.to_dict(), self.keys))



from serverless.service import YamlOrderedDict


class GenericArn(YamlOrderedDict):
    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_str(str(data))


class SQSArn(GenericArn):
    yaml_tag = "!Arn"

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"arn:aws:sqs:${{aws:region}}:${{aws:accountId}}:{self.name}"


class EventBridgeBusArn(GenericArn):
    yaml_tag = "!Arn"

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"arn:aws:events:${{aws:region}}:${{aws:accountId}}:event-bus/{self.name}"



from serverless.aws.resources.kms import EncryptableResource
from serverless.service.plugins.generic import Generic
from serverless.service.plugins.iam_roles import IAMRoles


class KMSGrant(Generic):
    def __init__(self, **kwds):
        super().__init__("serverless-kms-grants")

    def enable(self, service):
        super().enable(service)
        service.custom.kmsGrants = []

    def pre_render(self, service):
        if service.plugins.get(IAMRoles):
            for fn in service.functions.all():
                service.custom["kmsGrants"].append(
                    dict(
                        kmsKeyId=EncryptableResource.encryption_alias(),
                        roleName=fn.iam.role,
                    )
                )

        else:
            service.custom["kmsGrants"].append(
                dict(kmsKeyId=EncryptableResource.encryption_alias(), roleName=service.provider.iam.role)
            )



import re

from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class SecretsManagerReader(IAMPreset):
    def __init__(self, resource):
        super().__init__(resource)

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        resource = str(self.resource) if self.resource.endswith("-??????") else f"{self.resource}-??????"

        policy_builder.allow(
            permissions=["secretsmanager:GetSecretValue"],
            resources="arn:aws:secretsmanager:${aws:region}:${aws:accountId}:secret:" + resource,
            sid="SecretsManager" + Identifier(re.sub(r"\W", "", self.resource)).pascal,
        )



from serverless.aws.iam import IAMPreset, PolicyBuilder


class XRayService(IAMPreset):
    def __init__(self):
        super().__init__("*")

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            sid="Xray",
            permissions=[
                "xray:PutTraceSegments",
                "xray:PutTelemetryRecords",
                "xray:GetSamplingRules",
                "xray:GetSamplingTargets",
                "xray:GetSamplingStatisticSummaries",
            ],
            resources=["*"],
        )



from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class KinesisReader(IAMPreset):
    def __init__(self, stream):
        if not stream.startswith("arn:aws"):
            stream = "arn:aws:kinesis:${aws:region}:${aws:accountId}:stream/" + stream

        self.stream = stream

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "kinesis:DescribeStream",
                "kinesis:GetRecords",
                "kinesis:GetShardIterator",
                "kinesis:ListShards",
                "kinesis:ListStreams",
            ],
            resources=[self.stream],
            sid=sid or "KinesisStreamReader" + Identifier(self.stream, safe=True).pascal,
        )


class KinesisWriter(IAMPreset):
    def __init__(self, stream):
        if not stream.startswith("arn:aws"):
            stream = "arn:aws:kinesis:${aws:region}:${aws:accountId}:stream/" + stream

        self.stream = stream

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "kinesis:DescribeStream",
                "kinesis:PutRecord",
                "kinesis:PutRecords",
            ],
            resources=[self.stream],
            sid=sid or "KinesisStreamWriter" + Identifier(self.stream, safe=True).pascal,
        )



from serverless.service.plugins.generic import Generic


class VpcDiscovery(Generic):
    """
    Plugin homepage: https://www.npmjs.com/package/serverless-vpc-discovery
    Example of use
    ```python
    service.plugins.add(VpcDiscovery("main", subnets=[{"tagKey": "Name", "tagValues": ["private-a", "private-b"]}]))
    ```
    """

    yaml_tag = "!VpcDiscovery"

    def __init__(self, vpc_name, subnet_names=None, subnets=None, security_group_names=None, security_groups=None):
        super().__init__("serverless-vpc-discovery")

        self.vpcName = vpc_name
        if all((subnet_names, subnets)):
            raise Exception("You can use either subnet_names or subnets.")
        if all((security_group_names, security_groups)):
            raise Exception("You can use either security_group_names or security_groups.")

        if subnet_names:
            self.subnets = [{"tagKey": "Name", "tagValues": subnet_names}]
        if subnets:
            self.subnets = subnets

        if security_group_names:
            self.subnets = [{"tagKey": "Name", "tagValues": security_group_names}]
        if security_groups:
            self.securityGroups = security_groups

    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        service.custom.vpcDiscovery = export



from troposphere import Join
from troposphere.dynamodb import Table

from serverless.aws.iam import IAMPreset, PolicyBuilder


class DynamoDBReader(IAMPreset):
    resource: Table

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "dynamodb:GetItem",
                "dynamodb:Query",
                "dynamodb:Scan",
            ],
            resources=[
                self.resource.get_att("Arn").to_dict(),
                Join(delimiter="", values=[self.resource.get_att("Arn").to_dict(), "/index/*"]).to_dict(),
            ],
            sid=sid or self.resource.name + "Reader",
        )


class DynamoDBWriter(IAMPreset):
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "dynamodb:BatchWriteItem",
                "dynamodb:DeleteItem",
                "dynamodb:UpdateItem",
                "dynamodb:PutItem",
            ],
            resources=[self.resource.get_att("Arn").to_dict()],
            sid=sid or self.resource.name + "Writer",
        )


class DynamoDBWriteOnly(IAMPreset):
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "dynamodb:BatchWriteItem",
                "dynamodb:UpdateItem",
                "dynamodb:PutItem",
            ],
            resources=[self.resource.get_att("Arn").to_dict()],
            sid=sid or self.resource.name + "WriteOnly",
        )


class DynamoDBDelete(IAMPreset):
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "dynamodb:DeleteItem",
            ],
            resources=[self.resource.get_att("Arn").to_dict()],
            sid=sid or self.resource.name + "Delete",
        )


class DynamoDBFullAccess(IAMPreset):
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=[
                "dynamodb:GetItem",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:BatchWriteItem",
                "dynamodb:DeleteItem",
                "dynamodb:UpdateItem",
                "dynamodb:PutItem",
            ],
            resources=[self.resource.get_att("Arn").to_dict()],
            sid=sid or self.resource.name + "FullAccess",
        )



from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class RedshiftConnect(IAMPreset):
    def __init__(self, cluster, db_name, db_user):
        self.cluster = cluster
        self.db_name = db_name
        self.db_user = db_user

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=["redshift:GetClusterCredentials"],
            resources=[
                "arn:aws:redshift:${aws:region}:${aws:accountId}:cluster:" + self.cluster,
                "arn:aws:redshift:${aws:region}:${aws:accountId}:dbname:" + f"{self.cluster}/{self.db_name}",
                "arn:aws:redshift:${aws:region}:${aws:accountId}:dbuser:" + f"{self.cluster}/{self.db_user}",
            ],
            sid=sid or "RedshiftConnection" + Identifier(self.cluster).pascal,
        )


class RedshiftQuery(RedshiftConnect):
    def apply(self, policy_builder: PolicyBuilder, sid=None):
        super().apply(policy_builder, sid)
        policy_builder.allow(
            permissions=[
                "redshift-data:DescribeTable",
                "redshift-data:ExecuteStatement",
                "redshift-data:BatchExecuteStatement",
                "redshift-data:DescribeStatement",
                "redshift-data:GetStatementResult",
            ],
            resources=["*"],
            sid=sid or "RedshiftQuery" + Identifier(self.cluster).pascal,
        )



from troposphere.kms import Key, Alias

from serverless.service.plugins.kms import KMSGrant
from serverless.service.types import Feature


class Encryption(Feature):
    def __init__(self):
        super().__init__()
        self.key = Key(
            title="ServiceEncryptionKey",
            Description="Encryption Key for ${self:service}",
            Enabled=True,
            PendingWindowInDays=14,
            EnableKeyRotation=True,
            KeyPolicy={
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "Enable IAM User Permissions",
                        "Effect": "Allow",
                        "Principal": {"AWS": "arn:aws:iam::${aws:accountId}:root"},
                        "Action": "kms:*",
                        "Resource": "*",
                    },
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "logs.${aws:region}.amazonaws.com"},
                        "Action": [
                            "kms:Encrypt*",
                            "kms:Decrypt*",
                            "kms:ReEncrypt*",
                            "kms:GenerateDataKey*",
                            "kms:Describe*",
                        ],
                        "Resource": "*",
                        "Condition": {
                            "ArnLike": {
                                "kms:EncryptionContext:aws:logs:arn": "arn:aws:logs:${aws:region}:${aws:accountId}:log-group:/services/${self:service}/*"
                            }
                        },
                    },
                ],
            },
        )

        self.alias = Alias(
            "ServiceEncryptionKeyAlias", AliasName="alias/${self:service}-${sls:stage}", TargetKeyId=self.key.Ref()
        )

    def pre_render(self, service):
        super().pre_render(service)
        for fn in service.functions.all():
            self.key.KeyPolicy["Statement"].append({
                "Effect": "Allow",
                "Principal": {"Service": "logs.${aws:region}.amazonaws.com"},
                "Action": [
                    "kms:Encrypt*",
                    "kms:Decrypt*",
                    "kms:ReEncrypt*",
                    "kms:GenerateDataKey*",
                    "kms:Describe*",
                ],
                "Resource": "*",
                "Condition": {
                    "ArnLike": {
                        "kms:EncryptionContext:aws:logs:arn": "arn:aws:logs:${aws:region}:${aws:accountId}:log-group:/aws/lambda/" + fn.name.spinal
                    }
                },
            })

    def enable(self, service):
        service.resources.add(self.key)
        service.resources.add(self.alias)
        service.plugins.add(KMSGrant())



from serverless.aws.iam import IAMPreset, PolicyBuilder
from serverless.service import Identifier


class Publish(IAMPreset):
    def __init__(self, event_bus):
        if not event_bus.startswith("arn:aws"):
            event_bus = "arn:aws:events:${aws:region}:${aws:accountId}:event-bus/" + event_bus

        self.event_bus = event_bus

    def apply(self, policy_builder: PolicyBuilder, sid=None):
        policy_builder.allow(
            permissions=["events:PutEvents"],
            resources=[self.event_bus],
            sid=sid or "EventBridgePublisher" + Identifier(self.event_bus, safe=True).pascal,
        )



from serverless.aws.functions.generic import Function
from serverless.service.types import YamlOrderedDict


class RetryPolicy(YamlOrderedDict):
    yaml_tag = "RetryPolicy"

    def __init__(self, maximumRetryAttempts=2, maximumEventAge=3600):
        super().__init__()
        self.maximumEventAge = maximumEventAge
        self.maximumRetryAttempts = maximumRetryAttempts


class EventBridgeEvent(YamlOrderedDict):
    yaml_tag = "eventBridge"

    def __init__(self, eventBus, pattern=None, deadLetterQueueArn=None, retryPolicy=None):
        super().__init__()
        self.eventBus = eventBus
        self.pattern = pattern

        if deadLetterQueueArn:
            self.deadLetterQueueArn = deadLetterQueueArn

        # if retryPolicy:
        self.retryPolicy = retryPolicy.copy()


class EventBridgeFunction(Function):
    yaml_tag = "!EventBridgeFunction"

    def __init__(
        self,
        service,
        name,
        description,
        eventBus,
        pattern=None,
        deadLetterQueueArn=None,
        retryPolicy=None,
        handler=None,
        timeout=None,
        layers=None,
        use_dlq=True,
        use_async_dlq=True,
        **kwargs,
    ):
        super().__init__(
            service, name, description, handler, timeout, layers, use_dlq=use_dlq, use_async_dlq=use_async_dlq, **kwargs
        )
        self.trigger(EventBridgeEvent(eventBus, pattern, deadLetterQueueArn, retryPolicy))



from serverless.service.types import YamlOrderedDict


class Environment(YamlOrderedDict):
    yaml_tag = "Environment"

    def __init__(self, **envs):
        super().__init__()
        self.envs = envs or {}

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_dict(data.envs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['serverless',
 'serverless.aws',
 'serverless.aws.alerts',
 'serverless.aws.features',
 'serverless.aws.functions',
 'serverless.aws.iam',
 'serverless.aws.resources',
 'serverless.integration',
 'serverless.service',
 'serverless.service.plugins']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'inflection>=0.5.1,<0.6.0', 'troposphere>=3.2,<3.3']

setup_kwargs = {
    'name': 'serverless-builder',
    'version': '2.7.0',
    'description': 'Python interface to easily generate `serverless.yml`.',
    'long_description': '<h2 align="center">serverless-builder</h2>\n<p align="center">\n<a href="https://pypi.org/project/serverless-builder/"><img alt="PyPI" src="https://img.shields.io/pypi/v/serverless-builder"></a>\n<a href="https://pypi.org/project/serverless-builder/"><img alt="Python" src="https://img.shields.io/pypi/pyversions/serverless-builder.svg"></a>\n<a href="https://github.com/epsylabs/serverless-builder/blob/master/LICENSE"><img alt="License" src="https://img.shields.io/pypi/l/lucyna.svg"></a>\n</p>\n\nPython interface to easily generate [serverless.yml](https://www.serverless.com/) file.\n\nMassive thanks goes to [@dxd1](https://github.com/dxd1) for his original idea and implementation.\n\n## Why\n`serverless.yml` easily can become a massive file with hundreds of lines, even if you have only a couple of services.\n\nModifying even semi advanced project can be difficult. Adding plugins and configuration for it, making sure your endless resources are configured correctly, making sure that you follow best practices.\nIf you will multiple that by all microservices you have, it can be really painful.\n\n## How\n`serverless-builder` comes and tries to help you with those tasks.\n- plugin management (with autoconfiguration)\n- function factory (with some best practice hints (DLQ))\n- autoconfiguration of some provider specific features like AWS X-Ray\n- monitoring support (wip)\n- easy resource manipulation with troposphere lib https://github.com/cloudtools/troposphere (but if you want you can use old good dict)\n- easier IAM management with predefined permission sets\n- built-in support for any serverless attributes\n\n# Example\n\n```python\nfrom serverless.aws.functions.event_bridge import RetryPolicy\nfrom serverless.aws.functions.http import HTTPFunction\nfrom serverless import Service\nfrom serverless.provider import AWSProvider\nfrom serverless.aws.features import XRay\nfrom serverless.aws.iam.dynamodb import DynamoDBReader\nfrom serverless.plugins import ComposedVars, PythonRequirements, Prune\n\nfrom troposphere.dynamodb import Table, AttributeDefinition, KeySchema\n\nservice = Service(\n    "service-name",\n    "some dummy service",\n    AWSProvider()\n)\nservice.plugins.add(ComposedVars())\nservice.plugins.add(Prune())\nservice.plugins.add(PythonRequirements())\n\ntable = Table(\n    "TestTable",\n    BillingMode="PAY_PER_REQUEST",\n    AttributeDefinitions=[\n        AttributeDefinition(AttributeName="name", AttributeType="S")\n    ],\n    KeySchema=[KeySchema(AttributeName="name", KeyType="HASH")]\n)\n\nservice.enable(XRay())\nservice.provider.iam.apply(DynamoDBReader(table))\n\nservice.builder.function.generic("test", "description")\nservice.builder.function.http("test", "description", "/", HTTPFunction.POST)\n\n# Multiple events with different paths and/or methods can be set up for the same handler\n# This will add the same handler to all of these: POST /, POST /alias, PUT /, PUT /alias\nservice.builder.function.http("test", "description", ["/", "/alias"], ["POST", "PUT"], handler="shared.handler")\n\n# Context with pre-defined setup\nwith service.preset(\n    layers=[{"Ref": "PythonRequirementsLambdaLayer"}],\n    handler="test.handlers.custom_handler.handle"\n) as p:\n    p.http_get("test-list", "List all tests", "/")\n    p.http_get("test-get", "Get one test", "/{test_id}")\n\nevent_bridge_function = service.builder.function.event_bridge(\n    "event_bridge_function",\n    "sample event bridge function",\n    "epsy",\n    {"source": ["saas.external"]},\n)\n\nevent_bridge_function.use_delivery_dlq(RetryPolicy(5, 300))\nevent_bridge_function.use_async_dlq()\n\nservice.resources.add(table)\n\nservice.render()\n```\n',
    'author': 'Epsy',
    'author_email': 'engineering@epsyhealth.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/epsylabs/serverless-builder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from serverless.service.plugins.generic import Generic


class IAMRoles(Generic):
    yaml_tag = "!IAMRolesPlugin"

    def __init__(self):
        super().__init__("serverless-iam-roles-per-function")

    def enable(self, service):
        pass



from serverless.service.plugins.composed_vars import ComposedVars
from serverless.service.plugins.deployment_bucket import DeploymentBucket
from serverless.service.plugins.domain_manager import DomainManager
from serverless.service.plugins.generic import Generic
from serverless.service.plugins.localstack import Localstack
from serverless.service.plugins.prune import Prune
from serverless.service.plugins.python_requirements import PythonRequirements
from serverless.service.plugins.step_functions import StepFunctions
from serverless.service.plugins.vpc_discovery import VpcDiscovery



from serverless.service.plugins.generic import Generic


class StepFunctions(Generic):
    yaml_tag = "!StepFunctionsPlugin"

    def __init__(self):
        super().__init__("serverless-step-functions")



from serverless.service.plugins.generic import Generic


class Localstack(Generic):
    yaml_tag = "!LocalstackPlugin"

    def __init__(self, stages=["local"], host="http://localhost", edgePort=4566, autostart=True, networks=[], mountCode=False, sudo=False):
        super().__init__("serverless-localstack")
        self.stages = stages
        self.host = host
        self.edgePort = edgePort
        self.autostart = autostart
        self.networks = networks
        self.lambda_settings = dict(mountCode=mountCode)
        self.docker = dict(sudo= sudo)


    def enable(self, service):
        export = dict(self)
        export.pop("name", None)

        export["lambda"] = export.get("lambda_settings")
        export.pop("lambda_settings", None)

        service.custom.localstack = export
        service.custom.stages = {stage: {} for stage in self.stages}
